/* Whirl and Pinch, Earle F. Philhower III <earle@ziplabel.com>, March 12, 2003
 * 
 * The following routines were modified from Federico Mena-Quintero's
 * whirlpinch.c filter for The GIMP. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Below is the original header from Frederico's and Scott's plugin:
 * orig: Whirl and Pinch plug-in --- two common distortions in one place
 * orig: Copyright (C) 1997 Federico Mena Quintero
 * orig: federico@nuclecu.unam.mx
 * orig: Copyright (C) 1997 Scott Goehring
 * orig: scott@poverty.bloomington.in.us
 */

#include <windows.h>
#include <stdio.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "plugin.h"
#ifndef M_PI
#define M_PI       3.14159265358979323846
#endif
#ifndef M_PI_2
#define M_PI_2     1.57079632679489661923
#endif

#define guchar unsigned char
#define gint int
#define img_bpp 3
#define MAX(a,b) (((a)>(b))?(a):(b))

static inline int calc_undistorted_coords(double wx, double wy, double whirl,
					  double pinch, double *x, double *y);

static inline guchar bilinear(double x, double y, guchar *values);

double radius, radius2;
double cen_x, cen_y, scale_x, scale_y;
int sel_x2, sel_x1, sel_y2, sel_y1;
int sel_width, sel_height;
double xhsiz, yhsiz;
double pinch;
double wpRadius;


static inline void SafeGetPixel(int x, int y,
				MINIGRAPHIC *src, guchar pixel[3])
{
  if (x<0) x = 0;
  if (y<0) y = 0;
  if (x>=(int)src->width) x = src->width-1;
  if (y>=(int)src->height) y = src->height-1;
  pixel[0] = src->bits[y*src->lineWidth+x*3+0];
  pixel[1] = src->bits[y*src->lineWidth+x*3+1];
  pixel[2] = src->bits[y*src->lineWidth+x*3+2];
}


void ImageIt(int preview, int iWhirl, int iPinch, int iRadius,
	     MINIGRAPHIC *src, MINIGRAPHIC *dst)
{
  guchar          *top_row, *bot_row;
  guchar          *top_p, *bot_p;
  gint             row, col;
  guchar           pixel[4][4];
  guchar           values[4];
  double           whirl;
  double           cx, cy;
  int              ix, iy;
  int              i;

  pinch = iPinch/100.0;
  wpRadius = iRadius/50.0;

  sel_x1 = sel_y1 = 0;
  sel_x2 = src->width;
  sel_y2 = src->height;

  sel_width  = sel_x2 - sel_x1;
  sel_height = sel_y2 - sel_y1;

  cen_x = (double) (sel_x1 + sel_x2 - 1) / 2.0;
  cen_y = (double) (sel_y1 + sel_y2 - 1) / 2.0;

  xhsiz = (double) (sel_width - 1) / 2.0;
  yhsiz = (double) (sel_height - 1) / 2.0;

  if (xhsiz < yhsiz) {
    scale_x = yhsiz / xhsiz;
    scale_y = 1.0;
  } else if (xhsiz > yhsiz) {
    scale_x = 1.0;
    scale_y = xhsiz / yhsiz;
  } else {
    scale_x = 1.0;
    scale_y = 1.0;
  } /* else */

  radius = MAX(xhsiz, yhsiz);

  whirl   = (iWhirl*360.0/100.0) * M_PI / 180.0;
  radius2 = radius*radius * wpRadius;

  /* Initialize rows */
  top_row = (guchar*)GlobalAlloc(GPTR, 3 * sel_width);
  bot_row = (guchar*)GlobalAlloc(GPTR, 3 * sel_width);

  /* Initialize pixel region */

  for (row = sel_y1; row <= ((sel_y1 + sel_y2) / 2); row++) {
    top_p = top_row;
    bot_p = bot_row + img_bpp * (sel_width - 1);

    for (col = sel_x1; col < sel_x2; col++) {
      if (calc_undistorted_coords(col, row, whirl, pinch, &cx, &cy)) {
	/* We are inside the distortion area */

	/* Top */

	if (cx >= 0.0)
	  ix = (int) cx;
	else
	  ix = -((int) -cx + 1);

	if (cy >= 0.0)
	  iy = (int) cy;
	else
	  iy = -((int) -cy + 1);

	if (preview) {
	  SafeGetPixel(ix, iy, src, pixel[0]);
	  *top_p++ = pixel[0][0];
	  *top_p++ = pixel[0][1];
	  *top_p++ = pixel[0][2];
	} else {
	  SafeGetPixel(ix, iy, src, pixel[0]);
	  SafeGetPixel(ix+1, iy, src, pixel[1]);
	  SafeGetPixel(ix, iy+1, src, pixel[2]);
	  SafeGetPixel(ix+1, iy+1, src, pixel[3]);
	  for (i = 0; i < img_bpp; i++) {
	    values[0] = pixel[0][i];
	    values[1] = pixel[1][i];
	    values[2] = pixel[2][i];
	    values[3] = pixel[3][i];

	    *top_p++ = bilinear(cx, cy, values);
	  } /* for */
	}

	/* Bottom */

	cx = cen_x + (cen_x - cx);
	cy = cen_y + (cen_y - cy);

	if (cx >= 0.0)
	  ix = (int) cx;
	else
	  ix = -((int) -cx + 1);

	if (cy >= 0.0)
	  iy = (int) cy;
	else
	  iy = -((int) -cy + 1);


	if (preview) {
	  SafeGetPixel(ix, iy, src, pixel[0]);
	  *bot_p++ = pixel[0][0];
	  *bot_p++ = pixel[0][1];
	  *bot_p++ = pixel[0][2];
	} else {
	  SafeGetPixel(ix, iy, src, pixel[0]);
	  SafeGetPixel(ix+1, iy, src, pixel[1]);
	  SafeGetPixel(ix, iy+1, src, pixel[2]);
	  SafeGetPixel(ix+1, iy+1, src, pixel[3]);
	  for (i = 0; i < img_bpp; i++) {
	    values[0] = pixel[0][i];
	    values[1] = pixel[1][i];
	    values[2] = pixel[2][i];
	    values[3] = pixel[3][i];

	    *bot_p++ = bilinear(cx, cy, values);
	  } /* for */
	}

	bot_p -= 2 * img_bpp; /* We move backwards! */
      } else {
	/* We are outside the distortion area; just copy the source pixels */

	/* Top */
	SafeGetPixel(col, row, src, pixel[0]);
	for (i = 0; i < img_bpp; i++)
	  *top_p++ = pixel[0][i];

	/* Bottom */
	int r, c;
	r = (sel_y2 - 1) - (row - sel_y1);
	c = (sel_x2 - 1) - (col - sel_x1);
	SafeGetPixel(c, r, src, pixel[0]);

	for (i = 0; i < img_bpp; i++)
	  *bot_p++ = pixel[0][i];

	bot_p -= 2 * img_bpp; /* We move backwards! */
      } /* else */
    } /* for */

    /* Paint rows to image */

    memcpy(dst->bits + dst->lineWidth * row, top_row, sel_width*3);
    memcpy(dst->bits + dst->lineWidth *
	   ((sel_y2 - 1) - (row - sel_y1)), bot_row, sel_width*3);

  } /* for */

  GlobalFree(top_row);
  GlobalFree(bot_row);

} /* whirl_pinch */


static inline int
calc_undistorted_coords(double wx, double wy,
			double whirl, double pinch,
			double *x, double *y)
{
  double dx, dy;
  double d, factor;
  double dist;
  double ang, sina, cosa;
  double pinfact;
  int    inside;

  /* Distances to center, scaled */

  dx = (wx - cen_x) * scale_x;
  dy = (wy - cen_y) * scale_y;

  /* Distance^2 to center of *circle* (scaled ellipse) */

  d = dx * dx + dy * dy;

  /* If inside circle, then distort.  Else, just return the same position */

  inside = (d < radius2);

  if (inside) {
    if (wpRadius==0) wpRadius=0.00001;
    if (radius==0) radius=0.000001;
    dist = sqrt(d / wpRadius) / radius;

    /* Pinch */
    pinfact=sin(M_PI_2 * dist);
    if (pinfact==0) pinfact=0.00001;
    factor = pow(pinfact, -pinch);

    dx *= factor;
    dy *= factor;

    /* Whirl */

    factor = 1.0 - dist;

    ang = whirl * factor * factor;

    sina = sin(ang);
    cosa = cos(ang);

    if (scale_x==0) scale_x = 0.0000001;
    if (scale_y==0) scale_y = 0.0000001;

    *x = (cosa * dx - sina * dy) / scale_x + cen_x;
    *y = (sina * dx + cosa * dy) / scale_y + cen_y;
  } else {
    *x = wx;
    *y = wy;
  } /* else */

  return inside;
} /* calc_undistorted_coords */


/*****/

static inline guchar
bilinear(double x, double y, guchar *values)
{
  double m0, m1;

  x = fmod(x, 1.0);
  y = fmod(y, 1.0);

  if (x < 0.0)
    x += 1.0;

  if (y < 0.0)
    y += 1.0;

  m0 = (double) values[0] + x * ((double) values[1] - values[0]);
  m1 = (double) values[2] + x * ((double) values[3] - values[2]);

  return (guchar) (m0 + y * (m1 - m0));
} /* bilinear */


